#include "glwidget.h"

#include <vector>

#include "shader.h"
#include "camera.h"
#include "data.h"

#include <time.h>
#include <trkfileio.h>

#include <fstream>


// #include <QDebug>


Camera camera(glm::vec3(0.0f, 0.0f, 3.0f));     //! Virtual Camera that captures the scene.

string line_data_file_name;

int tab_index = 0;      //! Index of the currently selected tab in the UI (0 for line rendering, 1 for point rendering)


// Variables for Line Data rendering

bool renderHalos = true; //! Used to decide whether halos of lines should be rendered or not.

float strip_width = 0.01f;      //! Width of view-aligned Triangle-Strip used to draw the lines.
float minLineWidth = 0.10f;     //! Minimal Width of the black line drawn on the Triangle-Strip.
float maxLineWidth = 0.75f;     //! Maximal Width of the black line drawn on the Triangle-Strip.

// 1 for exponential, 2 for linear depth cueing
int depth_cueing = 1;   //! Parameter for selecting either exponential or linear depth-cueing.
int depth_falloff = 1;  //! Parameter for deciding how strong the Depth-Cueing should be.

int start_data = 0;     //! Index of the first Track that is loaded from the .trk Data-Set.
int end_data = 22796;   //! Index of the last Track that is loaded from the .trk Data-Set.

bool hair_thin_lines = false;

glm::mat4 model_line         = glm::mat4(1.0f);     //! Model Matrix for transforming the Line-Data to World Space
glm::mat4 view_line          = glm::mat4(1.0f);     //! View Matrix for transforming the Line-Data to Eye Space
glm::mat4 projection_line    = glm::mat4(1.0f);     //! Projection Matrix for transforming the Line-Data to HDC Space

glm::mat4 rotation_line    = glm::mat4(1.0f);       //! Rotation Component of the Model Matrix for Line Data
glm::mat4 translation_line    = glm::mat4(1.0f);    //! Translation Component of the Model Matrix for Line Data
glm::mat4 scale_line    = glm::mat4(1.0f);          //! Scale Component of the Model Matrix for Line Data

glm::mat4 initial_translation_line    = glm::mat4(1.0f);    //! Initial Translation that is applied to the Line-Data for Centering
glm::mat4 initial_scale_line    = glm::mat4(1.0f);          //! Initial Scaling that is applied to the Line-Data

std::vector<float> vertex_data_line;        //! Vertices of the View-Aligned Triangle Strips
std::vector<GLuint> indices_data_line;      //! Indices of the View-Aligned Triangle Strips
std::vector<float> directions_data_line;    //! Line Directions of the View-Aligned Triangle Strip
std::vector<float> uv_data_line;            //! UV-Coordinates of the View-Aligned Triangle Strip




// Variables for Point Data Rendering

bool renderHalos_point = true;  //! Used to decide whether halos of points should be rendered or not.

float cone_width = 0.05f;       //! Width of the quadratic Billboard the points are drawn on.
float point_radius = 0.40f;     //! Radius of the circle drawn onto the Billboard.

glm::mat4 model_point         = glm::mat4(1.0f); //! Model Matrix for transforming the Point-Data to World Space
glm::mat4 view_point          = glm::mat4(1.0f); //! View Matrix for transforming the Point-Data to Eye Space
glm::mat4 projection_point    = glm::mat4(1.0f); //! Projection Matrix for transforming the Point-Data to NDC Space

glm::mat4 rotation_point    = glm::mat4(1.0f);      //! Rotation Component of the Model Matrix for Point Data
glm::mat4 translation_point    = glm::mat4(1.0f);   //! Translation Component of the Model Matrix for Point Data
glm::mat4 scale_point    = glm::mat4(1.0f);         //! Scale Component of the Model Matrix for Point Data

glm::mat4 initial_translation_point    = glm::mat4(1.0f); //! Initial Translation that is applied to the Point-Data for Centering
glm::mat4 initial_scale_point    = glm::mat4(1.0f);       //! Initial Scale that is applied to the Point-Data

std::vector<float> vertex_data_point;       //! Vertices of the Billboards for rendering Points
std::vector<GLuint> indices_data_point;     //! Indices of the Billboards for rendering Points
std::vector<float> uv_data_point;           //! UV-Coordinates of the Billboards for rendering Points

/**
 * @brief import_line_data Reads Line-Data from a trk. File and generates Vertices, Indices, Directions and UV-Coordinates needed to construct the Triangle-Strips.
 * @param input_file_path The path to the .trk File that should be read.
 * @param start Index of the first Track that should be read.
 * @param end Index of the last Track that should be read.
 * @return Returns true if .trk File was successfully read, false else.
 */
bool import_line_data(string input_file_path, int start, int end);

/**
 * @brief duplicate_vertices Duplicates the vertices in the input-vector and adds them to {@link #vertex_data_line}
 * @param input_points Array of x-, y- and z-Coordinates of a vector.
 */
void duplicate_vertices(std::vector<float> input_points);

/**
 * @brief calculate_directions
 * @param input_points
 */
void calculate_directions(std::vector<float> input_points);

/**
 * @brief calculate_UVs
 * @param input_points
 */
void calculate_UVs(std::vector<float> input_points);

/**
 * @brief calculate_Indices
 * @param input_points
 */
void calculate_Indices(std::vector<float> input_points);

/**
 * @brief import_point_data Imports point-data from a .txt file and generates the Vertices, Indices and UV-Coordinates needed to render the Billboards.
 * @param input_file_path The .txt file that is to be imported.
 */
void import_point_data(string input_file_path);

/**
 * @brief outputVertex Writes the Coordinates of a three-dimensional vertex to the concole.
 * @param vertex The Vertex to be printed.
 */
void outputVertex(glm::vec3 vertex);

GLWidget::GLWidget(QWidget *parent) :
    QOpenGLWidget(parent),
    m_vertexBuffer_line(QOpenGLBuffer ::VertexBuffer),
    m_directionBuffer_line(QOpenGLBuffer ::VertexBuffer),
    m_uvBuffer_line(QOpenGLBuffer ::VertexBuffer),
    m_indexBuffer_line(QOpenGLBuffer ::IndexBuffer),

    m_vertexBuffer_point(QOpenGLBuffer ::VertexBuffer),
    m_uvBuffer_point(QOpenGLBuffer ::VertexBuffer),
    m_indexBuffer_point(QOpenGLBuffer ::IndexBuffer),

    texture(0)

{
    setFocusPolicy(Qt::StrongFocus);
}

GLWidget::~GLWidget()
{

}

void GLWidget::resizeGL(int w, int h)
{
    glViewport( 0, 0, w, qMax( h, 1 ) );
}











// ------------------------------------------ RENDERING ----------------------------------------------------



void GLWidget::paintGL()
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    // Render Line Data when in Tab 0
    if(tab_index == 0) {

        texture->bind();

        m_VAO_line.bind();
        m_vertexBuffer_line.bind();
        m_indexBuffer_line.bind();
        m_directionBuffer_line.bind();
        m_uvBuffer_line.bind();

        m_shader_line.bind();

        // Construct Model Matrix from rotation, translation and scale
        // Initial Rotation and Translation are applied beforehand to center the model and scale it in a meaningful way
        model_line = scale_line * translation_line * rotation_line * initial_scale_line * initial_translation_line * glm::mat4(1.0f);

        view_line = camera.GetViewMatrix();
        projection_line = glm::perspective(glm::radians(45.0f), (float)this->width() / (float)this->height(), 0.1f, 100.0f);


        m_shader_line.setUniformValue("model", QMatrix4x4(glm::value_ptr(model_line)).transposed());
        m_shader_line.setUniformValue("view", QMatrix4x4(glm::value_ptr(view_line)).transposed());
        m_shader_line.setUniformValue("projection", QMatrix4x4(glm::value_ptr(projection_line)).transposed());

        m_shader_line.setUniformValue("cameraPosition", QVector3D(camera.camera_position.x,camera.camera_position.y,camera.camera_position.z));

        m_shader_line.setUniformValue("renderHalos", renderHalos);
        m_shader_line.setUniformValue("stripWidth", strip_width);
        m_shader_line.setUniformValue("minRatio", minLineWidth);
        m_shader_line.setUniformValue("maxRatio", maxLineWidth);
        m_shader_line.setUniformValue("depth_cueing", depth_cueing);
        m_shader_line.setUniformValue("falloff", depth_falloff);

        glDrawElements(GL_TRIANGLES, indices_data_line.size(), GL_UNSIGNED_INT, 0);

        m_VAO_line.release();
        m_vertexBuffer_line.release();
        m_indexBuffer_line.release();
        m_directionBuffer_line.release();
        m_uvBuffer_line.release();
        texture->release();
    }
    else if(tab_index == 1) {
        m_VAO_point.bind();
        m_vertexBuffer_point.bind();
        m_indexBuffer_point.bind();
        m_uvBuffer_point.bind();

        m_shader_point.bind();

        // Construct Model Matrix from rotation, translation and scale
        // Initial Rotation and Translation are applied beforehand to center the model and scale it in a meaningful way
        model_point = scale_point * translation_point * rotation_point * initial_scale_point * initial_translation_point * glm::mat4(1.0f);

        view_point = camera.GetViewMatrix();
        projection_point = glm::perspective(glm::radians(45.0f), (float)this->width() / (float)this->height(), 0.1f, 100.0f);


        m_shader_point.setUniformValue("model", QMatrix4x4(glm::value_ptr(model_point)).transposed());
        m_shader_point.setUniformValue("view", QMatrix4x4(glm::value_ptr(view_point)).transposed());
        m_shader_point.setUniformValue("projection", QMatrix4x4(glm::value_ptr(projection_point)).transposed());

        m_shader_point.setUniformValue("cameraPosition", QVector3D(camera.camera_position.x,camera.camera_position.y,camera.camera_position.z));

        m_shader_point.setUniformValue("renderHalos", renderHalos_point);
        m_shader_point.setUniformValue("coneWidth", cone_width);
        m_shader_point.setUniformValue("radius", point_radius);
        m_shader_point.setUniformValue("maxRatio", maxLineWidth);
        m_shader_point.setUniformValue("depth_cueing", depth_cueing);
        m_shader_point.setUniformValue("falloff", depth_falloff);

        glDrawElements(GL_TRIANGLES, indices_data_point.size(), GL_UNSIGNED_INT, 0);

        m_VAO_point.release();
        m_vertexBuffer_point.release();
        m_indexBuffer_point.release();
        m_uvBuffer_point.release();
    }
}











// ------------------------------------------ INITIALIZE OPENGL, SHADERS, BUFFERS, TEXTURE ----------------------------------------------------


void GLWidget::initializeGL()
{

    std::cout << "------- Initializing Application -------" << std::endl;

    glClearColor(1.0, 1.0, 1.0, 1);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_MULTISAMPLE);

    // Activate for Wireframe Mode
    // glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );


    // Import Data
    if(!import_line_data("HCP842_tractography.trk", 0, 50000)){
        std::cout << "Line Data Import Failed" << std::endl;
    }
    line_data_file_name= "HCP842_tractography.trk";

    import_point_data("point_data_bunny.txt");


    // Prepare Shaders
    if (!prepareShaderProgram(m_shader_line, "Shader/line.vr", "Shader/line.fr"))
        return;

    if (!prepareShaderProgram(m_shader_point, "Shader/point.vr", "Shader/point.fr"))
        return;


    // Setup VAOs and Buffers
    setup_line_buffers();
    setup_point_buffers();

    initialize_texture();

    std::cout << "Setup Complete." << std::endl;

}

bool GLWidget::prepareShaderProgram(QOpenGLShaderProgram& shader,
                                    const QString& vertexShaderPath,
                                    const QString& fragmentShaderPath)
{

    std::cout << "Preparing Shaders..." << std::endl;

    // Loading the Vertex-Shader
    bool result = shader.addShaderFromSourceFile(QOpenGLShader::Vertex, vertexShaderPath);
    if (!result)
        qWarning() << shader.log();

    // Loading the Fragment-Shader
    result = shader.addShaderFromSourceFile(QOpenGLShader::Fragment, fragmentShaderPath);
    if (!result)
        qWarning() << shader.log();

    // Linking Vertex- and Fragment-Shader
    result = shader.link();
    if (!result)
        qWarning() << "Could not link shader program:" << shader.log();

    std::cout << "Shader Preparation Complete." << std::endl;

    return result;
}


void GLWidget::setup_line_buffers()
{

    std::cout << "Setting up OpenGL Buffers for Line Data..." << std::endl;

    m_VAO_line.create();
    m_VAO_line.bind();

    cout << "Creating Vertex Buffer..." << endl;

    m_vertexBuffer_line.create();
    m_vertexBuffer_line.setUsagePattern( QOpenGLBuffer::StaticDraw);
    m_vertexBuffer_line.bind();
    m_vertexBuffer_line.allocate( &vertex_data_line[0], vertex_data_line.size()*sizeof(float) );

    cout << "Creating Index Buffer..." << endl;

    m_indexBuffer_line.create();
    m_indexBuffer_line.setUsagePattern(QOpenGLBuffer::StaticDraw);
    m_indexBuffer_line.bind();
    m_indexBuffer_line.allocate(&indices_data_line[0], indices_data_line.size()*sizeof(GLuint));

    cout << "Creating Direction Buffer..." << endl;

    m_directionBuffer_line.create();
    m_directionBuffer_line.setUsagePattern(QOpenGLBuffer::StaticDraw);
    m_directionBuffer_line.bind();
    m_directionBuffer_line.allocate(&directions_data_line[0], directions_data_line.size()*sizeof(float) );

    cout << "Creating UV Buffer..." << endl;

    m_uvBuffer_line.create();
    m_uvBuffer_line.setUsagePattern(QOpenGLBuffer::StaticDraw);
    m_uvBuffer_line.bind();
    m_uvBuffer_line.allocate(&uv_data_line[0], uv_data_line.size()*sizeof(float) );

    cout << "Buffer Creation Complete." << endl;

    m_shader_line.bind();

    cout << "Sending Data To Shaders..." << endl;

    m_vertexBuffer_line.bind();
    m_shader_line.setAttributeBuffer( "vertexPosition", GL_FLOAT, 0, 3 );
    m_shader_line.enableAttributeArray( "vertexPosition" );

    m_directionBuffer_line.bind();
    m_shader_line.setAttributeBuffer( "vertexDirection", GL_FLOAT, 0, 3 );
    m_shader_line.enableAttributeArray( "vertexDirection" );

    m_uvBuffer_line.bind();
    m_shader_line.setAttributeBuffer( "vertexUVs", GL_FLOAT, 0, 2 );
    m_shader_line.enableAttributeArray( "vertexUVs" );

    cout << "Transfer Complete." << endl;

    cout << "Releasing Objects." << endl;

    m_VAO_line.release();
    m_vertexBuffer_line.release();
    m_indexBuffer_line.release();
    m_directionBuffer_line.release();
    m_uvBuffer_line.release();

    cout << "Line Data Buffer Setup Complete." << endl;

}


void GLWidget::setup_point_buffers()
{

    std::cout << "Setting up OpenGL Buffers for Point Data..." << std::endl;

    m_VAO_point.create();
    m_VAO_point.bind();

    cout << "Creating Vertex Buffer..." << endl;

    m_vertexBuffer_point.create();
    m_vertexBuffer_point.setUsagePattern( QOpenGLBuffer::StaticDraw);
    m_vertexBuffer_point.bind();
    m_vertexBuffer_point.allocate( &vertex_data_point[0], vertex_data_point.size()*sizeof(float) );

    cout << "Creating Index Buffer..." << endl;

    m_indexBuffer_point.create();
    m_indexBuffer_point.setUsagePattern(QOpenGLBuffer::StaticDraw);
    m_indexBuffer_point.bind();
    m_indexBuffer_point.allocate(&indices_data_point[0], indices_data_point.size()*sizeof(GLuint));

    cout << "Creating UV Buffer..." << endl;

    m_uvBuffer_point.create();
    m_uvBuffer_point.setUsagePattern(QOpenGLBuffer::StaticDraw);
    m_uvBuffer_point.bind();
    m_uvBuffer_point.allocate(&uv_data_point[0], uv_data_point.size()*sizeof(float) );

    cout << "Buffer Creation Complete." << endl;

    m_shader_point.bind();

    cout << "Sending Data To Shaders..." << endl;

    m_vertexBuffer_point.bind();
    m_shader_point.setAttributeBuffer( "vertexPosition", GL_FLOAT, 0, 3 );
    m_shader_point.enableAttributeArray( "vertexPosition" );

    m_uvBuffer_point.bind();
    m_shader_point.setAttributeBuffer( "vertexUVs", GL_FLOAT, 0, 2 );
    m_shader_point.enableAttributeArray( "vertexUVs" );

    cout << "Transfer Complete." << endl;

    cout << "Releasing Objects." << endl;

    m_VAO_point.release();
    m_vertexBuffer_point.release();
    m_indexBuffer_point.release();
    m_uvBuffer_point.release();

    cout << "Point Data Buffer Setup Complete." << endl;

}



void GLWidget::initialize_texture()
{
    texture = new QOpenGLTexture(QImage("Texture/mask.png").mirrored());

    texture->setMinificationFilter(QOpenGLTexture::Nearest);
    texture->setMagnificationFilter(QOpenGLTexture::Linear);

    texture->setWrapMode(QOpenGLTexture::ClampToBorder);

    texture->generateMipMaps();
}





// ------------------------------------------ DATA IMPORT ----------------------------------------------------




bool import_line_data(string strInputFilePath, int start, int end)
{

    std::cout << "Importing Data..." << std::endl;

    // string strOutputFilePath = "TESTTR.trk";

    /// create reader and open file
    TrkFileReader cReader(strInputFilePath);

    if (!cReader.open())
        return false;

    vector<float> cTrk;
    size_t iTotalTrackNum = cReader.getTotalTrkNum();  /// total number of tracks in input file
    // srand((unsigned int)(time(nullptr)));                           /// random seed


    // The center of the data-set is calculated for centering the model at (0/0/0)
    float center_x = 0.0f;
    float center_y = 0.0f;
    float center_z = 0.0f;
    int num_vertices = 0;

    cout << "Number of Tracks to Import: " << iTotalTrackNum << endl;
    for (size_t i = 0; i < iTotalTrackNum; i++)
    {

        // load all Tracks between Start-Index and End-Index
        if(i > start && i < end ) {

            //if (rand() % 1000 < 500)   /// 0.1% random sampling
            //    continue;

            cTrk.clear();
            bool success = cReader.readTrack(i, cTrk); /// read one track from input

            for (int i =0; i<cTrk.size(); i+=3) {
                center_x += cTrk[i];
                center_y += cTrk[i+1];
                center_z += cTrk[i+2];
                num_vertices++;
            }

            int minLength = 2;

            if(success == true && cTrk.size() >= minLength*3) {
                duplicate_vertices(cTrk);
                calculate_directions(cTrk);
                calculate_UVs(cTrk);
                calculate_Indices(cTrk);
            }
            else {
                // cout << "Failed to read Track " << i << endl;
            }
        }

    }

    // Calculate Center of Model
    center_x = center_x / num_vertices;
    center_y = center_y / num_vertices;
    center_z = center_z / num_vertices;
    std::cout << "Center: " << center_x << " / " << center_y << " / " << center_z << std::endl;

    // Translate the model such that its center is at (0/0/0) and scale in a meaningful way
    float scale_value = 0.01f;
    initial_scale_line = glm::scale(scale_line, glm::vec3(scale_value,scale_value,scale_value));
    initial_translation_line = glm::translate(translation_line, glm::vec3(-center_x, -center_y, -center_z));

    /// close input and output file
    cReader.close();

    std::cout << "Data Import Complete." << std::endl;

    return true;
}


void import_point_data(string input_file_path) {
    std::ifstream infile(input_file_path);

    std::cout << "Importing Point Data" << std::endl;

    int indices = 0;

    // The center of the data-set is calculated for centering the model at (0/0/0)
    float center_x = 0.0f;
    float center_y = 0.0f;
    float center_z = 0.0f;
    int num_vertices = 0;

    float a, b, c;
    while (infile >> a >> b >> c)
    {

        if(indices > 3) {
            // break;
        }

        // Adds each vertex to the vertex buffer 4 times to produce 4 corners of the billboards
        for(int i = 0; i <= 3; i++) {
            vertex_data_point.push_back(a);
            vertex_data_point.push_back(b);
            vertex_data_point.push_back(c);
        }

        // Generate UV-Coordinates for the Corner-Points ([0,0] [1,0] [1,1] and [0,1])
        uv_data_point.push_back(0);
        uv_data_point.push_back(0);
        uv_data_point.push_back(1);
        uv_data_point.push_back(0);
        uv_data_point.push_back(1);
        uv_data_point.push_back(1);
        uv_data_point.push_back(0);
        uv_data_point.push_back(1);

        // Generate Indices of the Triangles that make up the Billboard ((0, 1, 2) and (0, 2, 3))
        indices_data_point.push_back(indices);
        indices_data_point.push_back(indices+1);
        indices_data_point.push_back(indices+2);
        indices_data_point.push_back(indices);
        indices_data_point.push_back(indices+2);
        indices_data_point.push_back(indices+3);

        // cout << "Triangle 1 " << ": " << indices << " / " << indices+1 << " / " << indices+2 << endl;
        // cout << "Triangle 2 " << ": " << indices << " / " << indices +2 << " / " << indices +3 << endl;

        center_x += a;
        center_y += b;
        center_z += c;
        num_vertices++;

        indices += 4;

    }

    // Calculate the Center of the model
    center_x = center_x / num_vertices;
    center_y = center_y / num_vertices;
    center_z = center_z / num_vertices;

    // center the model at (0/0/0) and scale in a meaningful way
    float scale_value = 10.0f;
    initial_scale_point = glm::scale(scale_point, glm::vec3(scale_value,scale_value,scale_value));
    initial_translation_point = glm::translate(translation_point, glm::vec3(-center_x, -center_y, -center_z));

    std::cout << "Point Data Import Finished" << std::endl;

}



void duplicate_vertices(std::vector<float> input_points)
{
    //Add 3D point to vertex data twice
    for (int i =0; i<input_points.size(); i+=3){
        for(int dupl = 0; dupl < 2; dupl++) {
            vertex_data_line.push_back(input_points[i]);
            vertex_data_line.push_back(input_points[i+1]);
            vertex_data_line.push_back(input_points[i+2]);
        }
    }
}

void calculate_directions(std::vector<float> input_points)
{
    // Tracks of length 1 do not have a direction
    if(input_points.size() <= 9) {
        directions_data_line.push_back(0.0);
        directions_data_line.push_back(0.0);
        directions_data_line.push_back(0.0);
        return;
    }


    for(int i =0; i<input_points.size(); i+=3){

        glm::vec3 vertex_direction;

        //For the frist vertex the direction is calculated normally by vector substraction
        if(i==0)
        {
            glm::vec3 current_vertex = glm::vec3(input_points[i], input_points[i+1], input_points[i+2]);
            glm::vec3 next_vertex = glm::vec3(input_points[i+3], input_points[i+4], input_points[i+5]);

            vertex_direction = next_vertex - current_vertex;
        }
        //For the last vertex the direction is also calculated normally by vector substraction
        else if(i+3 == input_points.size())
        {
            glm::vec3 current_vertex = glm::vec3(input_points[i], input_points[i+1], input_points[i+2]);
            glm::vec3 previous_vertex = glm::vec3(input_points[i-3], input_points[i-2], input_points[i-1]);

            vertex_direction = current_vertex - previous_vertex;
        }
        //For intermediate vertices, the normalized average direction of both line segments adjacent to the vertex is taken
        else{
            glm::vec3 previous_vertex = glm::vec3(input_points[i-3], input_points[i-2], input_points[i-1]);
            glm::vec3 current_vertex = glm::vec3(input_points[i], input_points[i+1], input_points[i+2]);
            glm::vec3 next_vertex = glm::vec3(input_points[i+3], input_points[i+4], input_points[i+5]);

            vertex_direction = ((current_vertex - previous_vertex)+(next_vertex - current_vertex)) / 2.f;
        }

        //Add direction twice for duplicated vertices
        directions_data_line.push_back(vertex_direction[0]);
        directions_data_line.push_back(vertex_direction[1]);
        directions_data_line.push_back(vertex_direction[2]);

        directions_data_line.push_back(vertex_direction[0]);
        directions_data_line.push_back(vertex_direction[1]);
        directions_data_line.push_back(vertex_direction[2]);
    }
}

void calculate_UVs(std::vector<float> input_points)
{
    int arraysize = 2*input_points.size()/3*2;

    for (int i =0; i<arraysize; i+=4){

        if(!hair_thin_lines){
            //TAPERING the first two line segments (non-linearly)
            if(i==0)
            {
                //First Vertex
                uv_data_line.push_back(0.2);
                uv_data_line.push_back(1); //left side of the strip

                //Duplicated Vertex
                uv_data_line.push_back(0.2);
                uv_data_line.push_back(0); //right side of the strip
            }
            else if (i==4)
            {
                //First Vertex
                uv_data_line.push_back(0.5);
                uv_data_line.push_back(1); //left side of the strip

                //Duplicated Vertex
                uv_data_line.push_back(0.5);
                uv_data_line.push_back(0); //right side of the strip
            }


            //TAPERING the last two line segments (non-linearly)
            else if (i==arraysize-8)
            {
                //First Vertex
                uv_data_line.push_back(0.5);
                uv_data_line.push_back(1); //left side of the strip

                //Duplicated Vertex
                uv_data_line.push_back(0.5);
                uv_data_line.push_back(0); //right side of the strip
            }
            else if (i==arraysize-4)
            {
                //First Vertex
                uv_data_line.push_back(0.2);
                uv_data_line.push_back(1); //left side of the strip

                //Duplicated Vertex
                uv_data_line.push_back(0.2);
                uv_data_line.push_back(0); //right side of the strip
            }


            //Intermediate line segments > No TAPERING
            else
            {
                //First Vertex
                uv_data_line.push_back(1);
                uv_data_line.push_back(1); //left side of the strip

                //Duplicated Vertex
                uv_data_line.push_back(1);
                uv_data_line.push_back(0); //right side of the strip
            }

        }else{

            //Lines are drawn very thin

            //First Vertex
            uv_data_line.push_back(0);
            uv_data_line.push_back(1); //left side of the strip

            //Duplicated Vertex
            uv_data_line.push_back(0);
            uv_data_line.push_back(0); //right side of the strip
        }

    }
}

void calculate_Indices(std::vector<float> input_points)
{

    //Because all vertices for all line tracks are stored in a single array, we need to have a stride betweed the each new track.
    GLuint initial_array_size=0;

    if(indices_data_line.size()>0){
        initial_array_size=indices_data_line[indices_data_line.size()-1]+1;
    }

    int num_vertices = input_points.size()/3;
    int num_triangles = 3*(num_vertices-1)*2;

    GLuint second_num = 0;
    //Setting up indices data using the previously defined stride
    for (GLuint i =0; i<num_triangles; i+=3){
        if(i==0){

            indices_data_line.push_back(i + initial_array_size);
            indices_data_line.push_back(i+1 +initial_array_size);
            indices_data_line.push_back(i+2 +initial_array_size);

            second_num = i+1;
        }else{

            indices_data_line.push_back(second_num  +initial_array_size);
            indices_data_line.push_back(second_num+1 +initial_array_size);
            indices_data_line.push_back(second_num+2 +initial_array_size);

            second_num += 1;
        }
    }
}








// ------------------------------------------ INPUT HANDLING ----------------------------------------------------



// ------------------------------------------ Mouse and Keyboard


glm::vec2 last_mouse_pos = glm::vec2(0.0f);     //! Position of the mouse in the last frame.
glm::vec2 mouse_dist = glm::vec2(0.0f);         //! Distance the mouse has travelled in x- and y-direction since last frame.
bool mouse_clicked = false;                     //! Indicates whether the mouse-button is held down or released.

void GLWidget::mousePressEvent(QMouseEvent * event)
{
    last_mouse_pos = glm::vec2(event->x(), event->y());
    mouse_clicked = true;

}

void GLWidget::mouseMoveEvent(QMouseEvent *event) {

    glm::vec2 current_mouse_pos = glm::vec2(event->x(), event->y());
    mouse_dist = current_mouse_pos - last_mouse_pos;

    // rotate the model if left mouse button is pressed
    if (event->buttons() & Qt::LeftButton) {

        if(tab_index == 0) {
            rotation_line = glm::rotate(glm::mat4(1.0f), (float)mouse_dist.x*0.01f, glm::vec3(0.0f, 1.0f, 0.0f)) * rotation_line;
            rotation_line = glm::rotate(glm::mat4(1.0f), (float)mouse_dist.y*0.01f, glm::vec3(1.0f, 0.0f, 0.0f)) * rotation_line;
        }
        else if(tab_index == 1) {
            rotation_point = glm::rotate(glm::mat4(1.0f), (float)mouse_dist.x*0.01f, glm::vec3(0.0f, 1.0f, 0.0f)) * rotation_point;
            rotation_point = glm::rotate(glm::mat4(1.0f), (float)mouse_dist.y*0.01f, glm::vec3(1.0f, 0.0f, 0.0f)) * rotation_point;
        }

        last_mouse_pos = current_mouse_pos;
    }

    // translate the model if middle mouse button is pressed
    else if(event->buttons() & Qt::MiddleButton) {
        if(tab_index == 0) {
            translation_line = glm::translate(glm::mat4(1.0f), glm::vec3((float)mouse_dist.x*0.005f, (float)-mouse_dist.y*0.005f, 0.0f)) * translation_line;
        }
        else if(tab_index == 1) {
            translation_point = glm::translate(glm::mat4(1.0f), glm::vec3((float)mouse_dist.x*0.005f, (float)-mouse_dist.y*0.005f, 0.0f)) * translation_point;
        }

    }

    last_mouse_pos = current_mouse_pos;

    update();

}

void GLWidget::wheelEvent(QWheelEvent * event) {

    float scale_factor = 0.001f * event->delta();
    // scale = glm::scale(scale, glm::vec3(scale_factor, scale_factor, scale_factor));
    camera.move_camera(FORWARD, scale_factor);
    update();
}

void GLWidget::keyPressEvent(QKeyEvent *event)
{

    switch (event->key())
    {
    case Qt::Key_W :
        camera.move_camera(FORWARD, 0.005f);
        break;
    case Qt::Key_S :
        camera.move_camera(BACKWARD, 0.005f);
        break;
    case Qt::Key_A :
        camera.move_camera(LEFT, 0.005f);
        break;
    case Qt::Key_D :
        camera.move_camera(RIGHT, 0.005f);
        break;
    default:
        break;
    }

    update();
}



// ------------------------------------------ UI Elements


void GLWidget::tabChanged(int value) {
    tab_index = value;
    std::cout << "Tab Changed to " << value << std::endl;
    update();
}



// UI Elements Line Data

void GLWidget::selectLineDataFile(){
    QString filename= QFileDialog::getOpenFileName(this, tr("Open File"), "C://", "All files (*.*);; Trk File (*.trk)");
    std::cout << filename.toUtf8().constData() << std::endl;

    line_data_file_name = filename.toUtf8().constData();

    /*vertex_data_line.clear();
    indices_data_line.clear();
    directions_data_line.clear();
    uv_data_line.clear();
    import_line_data(filename.toUtf8().constData(), start_data,end_data);
    setup_line_buffers();

    update();*/
    reloadData();
}

void GLWidget::resetView() {

    // reset components of model-matrix
    rotation_line    = glm::mat4(1.0f);
    translation_line    = glm::mat4(1.0f);
    scale_line    = glm::mat4(1.0f);

    // reset camera to initial position
    camera.camera_position = glm::vec3(0.0f, 0.0f, 3.0f);

    std::cout << "Line Data View Reset" << std::endl;
    update();
}

void GLWidget::renderHalosChanged() {
    renderHalos = !renderHalos;
    std::cout << "Rendering Halos: " << renderHalos << std::endl;
    update();
}

void GLWidget::stripWidthChanged(int value) {
    float base_strip_width = 0.01f;

    strip_width = base_strip_width * (value/10.0);

    std::cout << "Strip Width changed to " << value << std::endl;
    update();

}

void GLWidget::minLineWidthChanged(int value) {
    minLineWidth = (value/100.f);

    if(minLineWidth > maxLineWidth) {
        adjustMaxLineWidthSlider(value);
        maxLineWidthChanged(value);
    }

    std::cout << "Min. Line Width: " << value << std::endl;
    update();
}

void GLWidget::maxLineWidthChanged(int value) {
    maxLineWidth = (value/100.f);

    if(maxLineWidth < minLineWidth) {
        adjustMinLineWidthSlider(value);
        minLineWidthChanged(value);
    }

    std::cout << "Max. Line Width: " << value << std::endl;
    update();
}

void GLWidget::expDepthCueingChecked() {
    depth_cueing = 1;
    std::cout << "Depth Cueing: exponential" << std::endl;
    update();
}

void GLWidget::linDepthCueingChecked() {
    depth_cueing = 2;
    std::cout << "Depth Cueing: linear" << std::endl;
    update();
}

void GLWidget::depthFalloffChanged(int value) {
    depth_falloff = value;
    std::cout << "Depth Falloff: " << value << std::endl;
    update();
}

void GLWidget::startDataChanged(int value) {
    start_data = value;
}

void GLWidget::endDataChanged(int value) {
    end_data = value;
}

void GLWidget::hairthinLinesExperiment() {
    hair_thin_lines = !hair_thin_lines;
    std::cout << "Draw hair thin lines: " << hair_thin_lines << std::endl;
    update();
}

void GLWidget::reloadData() {
    vertex_data_line.clear();
    indices_data_line.clear();
    directions_data_line.clear();
    uv_data_line.clear();
    import_line_data(line_data_file_name, start_data,end_data);
    setup_line_buffers();

    std::cout << "Reloading Tracks from " << start_data << " to " << end_data << std::endl;
    update();

}



// UI Elements Point Data

void GLWidget::selectPointDataFile(){
    QString filename= QFileDialog::getOpenFileName(this, tr("Open File"), "C://", "All files (*.*);; Txt File (*.txt)");
    std::cout << filename.toUtf8().constData() << std::endl;

    vertex_data_point.clear();
    indices_data_point.clear();
    uv_data_point.clear();
    import_point_data(filename.toUtf8().constData());

    setup_point_buffers();
    update();

}

void GLWidget::resetView_Point() {
    // reset components of model-matrix
    rotation_point    = glm::mat4(1.0f);
    translation_point    = glm::mat4(1.0f);
    scale_point    = glm::mat4(1.0f);

    // reset camera to initial position
    camera.camera_position = glm::vec3(0.0f, 0.0f, 3.0f);

    std::cout << "Point Data View Reset" << std::endl;
    update();
}

void GLWidget::renderHalosChanged_Point() {
    renderHalos_point = !renderHalos_point;
    std::cout << "Rendering Point Halos: " << renderHalos_point << std::endl;
    update();
}

void GLWidget::coneSizeChanged(int value) {
    float base_cone_width = 0.01f;

    cone_width = base_cone_width * (value/10.0);

    std::cout << "Cone Width changed to " << value << std::endl;
    update();
}

void GLWidget::pointDiameterChanged(int value) {
    point_radius = (value/100.f);

    std::cout << "Point Radius: " << value << std::endl;
    update();
}









// ------------------------------------------ HELPER FUNCTIONS ----------------------------------------------------

void outputVertex(glm::vec3 vertex) {
    std::cout << vertex.x << " / " << vertex.y << " / " << vertex.z << std::endl;
}
