#ifndef GLWIDGET_H
#define GLWIDGET_H

#include <glew.h>
#include <QWidget>
#include <qopenglwidget.h>
#include <QOpenGLShaderProgram>
#include <QOpenGLVertexArrayObject>
#include <QOpenGLBuffer>
#include <QOpenGLTexture>


#include <glm.hpp>
#include <gtc/matrix_access.hpp>
#include <gtc/type_ptr.hpp>

#include <stdlib.h>

#include <QDateTime>
#include <QKeyEvent>

#include <QFileDialog>




class GLWidget : public QOpenGLWidget
{
    Q_OBJECT
public:
    explicit GLWidget(QWidget *parent = 0);
    ~GLWidget();

    /**
     * @brief initializeGL Initializes the application by loading the data, preparing the shaders and setting up VAOs and Buffers.
     */
    void initializeGL();

    /**
     * @brief paintGL Renders the line- and point-data to the GUI.
     */
    void paintGL();

    /**
     * @brief resizeGL Resizes the GUI.
     * @param w New width of the window.
     * @param h New height of the window.
     */
    void resizeGL(int w, int h);

    /**
     * @brief setup_line_buffers Creates VAO and Buffers needed to correctly render the line-data as view-aligned Triangle-Strips
     */
    void setup_line_buffers();

    /**
     * @brief setup_point_buffers Creates VAO and Buffers needed to correctly render the point-data as view-aligned Billboards
     */
    void setup_point_buffers();

    /**
     * @brief Load texuture mask that is used for tapering. Additionally, apply filtering and generate mipmapping
     */
    void initialize_texture();

    void keyPressEvent( QKeyEvent *);
    void mousePressEvent(QMouseEvent * event);
    void mouseMoveEvent(QMouseEvent *event);
    void wheelEvent(QWheelEvent * event);

// Slots are functions that can be called by elements in the UI
public slots:

    void tabChanged(int value);

    // UI Slots Line Data
    void selectLineDataFile();
    void resetView();
    void renderHalosChanged();
    void stripWidthChanged(int value);
    void minLineWidthChanged(int value);
    void maxLineWidthChanged(int value);
    void expDepthCueingChecked();
    void linDepthCueingChecked();
    void depthFalloffChanged(int value);
    void startDataChanged(int value);
    void endDataChanged(int value);
    void hairthinLinesExperiment();
    void reloadData();

    // UI Slots Point Data
    void selectPointDataFile();
    void resetView_Point();
    void renderHalosChanged_Point();
    void coneSizeChanged(int value);
    void pointDiameterChanged(int value);


// Signals are methods that are called to trigger functions in the UI
signals:
    void adjustMaxLineWidthSlider(int newValue);
    void adjustMinLineWidthSlider(int newValue);
    void dataLoadProgress(int value);

private:
    /**
     * @brief prepareShaderProgram Creates and Links Vertex- and Fragment-Shader
     * @param shader Reference to the shader that is being created.
     * @param vertexShaderPath Path to the Vertex-Shader.
     * @param fragmentShaderPath  Path to the Fragment-Shader.
     * @return
     */
    bool prepareShaderProgram(QOpenGLShaderProgram& shader,
                              const QString& vertexShaderPath,
                              const QString& fragmentShaderPath);
    QOpenGLShaderProgram  m_shader_line;    //! Shader Program for rendering Line Data.
    QOpenGLVertexArrayObject m_VAO_line;    //! VAO for rendering Line Data.
    QOpenGLBuffer  m_vertexBuffer_line;     //! Contains the Vertices of the view-aligned Triangle-Strips.
    QOpenGLBuffer  m_directionBuffer_line;  //! Contains the Directions of the view-aligned Triangle-Strips.
    QOpenGLBuffer  m_uvBuffer_line;         //! Contains the UV-Coordinates of the vertices of the view-aligned Triangle-Strips.
    QOpenGLBuffer  m_indexBuffer_line;      //! Contains the Indices of the triangles in the view-aligned Triangle-Strips.

    QOpenGLShaderProgram  m_shader_point;   //! Shader Program for rendering Point Data.
    QOpenGLVertexArrayObject m_VAO_point;   //! VAO for rendering Point Data.
    QOpenGLBuffer  m_vertexBuffer_point;    //! Contains the Vertices of the view-aligned Billboards.
    QOpenGLBuffer  m_uvBuffer_point;        //! Contains the UV-coordinates of the Vertices of the view-aligned Triangle-Strips.
    QOpenGLBuffer  m_indexBuffer_point;     //! Contains the Indices of the Triangles that make up the Billboards.

    QOpenGLTexture *texture;                //! Contains texture mask for tapering

};

#endif // GLWIDGET_H
